#!/usr/bin/env python3
"""
2020/01/29
    an abstract class for pegasus workflows that work on bioinformatic data
"""
import os, sys
import logging
from palos import PassingData
from . AbstractWorkflow import AbstractWorkflow

ParentClass = AbstractWorkflow
class AbstractBioinfoWorkflow(ParentClass):
    __doc__ = __doc__
    option_default_dict = ParentClass.option_default_dict.copy()
    option_default_dict.update({
        ("plinkPath", 1, ): ["bin/plink", '', 1, 
            'path to plink, http://pngu.mgh.harvard.edu/~purcell/plink/index.shtml'],
        })
    def __init__(self,
        input_path=None,
        inputSuffixList=None, 
        pegasusFolderName='folder', output_path=None,
        
        tmpDir='/tmp/', max_walltime=4320,
        home_path=None, javaPath=None, 
        pymodulePath="src/pymodule", thisModulePath=None,
        jvmVirtualByPhysicalMemoryRatio=1.2,

        site_handler='condor', input_site_handler='condor', cluster_size=30,
        needSSHDBTunnel=False, commit=False,
        debug=False, report=False):
        """
        20200129
        """
        ParentClass.__init__(self,
            input_path=input_path,
            inputSuffixList=inputSuffixList,
            pegasusFolderName=pegasusFolderName,
            output_path=output_path,
            
            tmpDir=tmpDir, max_walltime=max_walltime, 
            home_path=home_path,
            javaPath=javaPath,
            pymodulePath=pymodulePath, thisModulePath=thisModulePath,
            jvmVirtualByPhysicalMemoryRatio=jvmVirtualByPhysicalMemoryRatio,
            
            site_handler=site_handler,
            input_site_handler=input_site_handler,
            cluster_size=cluster_size,

            needSSHDBTunnel=needSSHDBTunnel, commit=commit,
            debug=debug, report=report)
    
    def registerBlastNucleotideDatabaseFile(self, ntDatabaseFname=None,  folderName=""):
        """
        moved from BlastWorkflow.py
        """
        return self.registerRefFastaFile(refFastaFname=ntDatabaseFname,
            registerAffiliateFiles=True,
            checkAffiliateFileExistence=True, addPicardDictFile=False,
            affiliateFilenameSuffixLs=['nin', 'nhr', 'nsq'],\
            folderName=folderName)

    def registerRefFastaFile(self, refFastaFname=None,
        registerAffiliateFiles=True,
        checkAffiliateFileExistence=True, addPicardDictFile=True,\
        affiliateFilenameSuffixLs=\
            ['fai', 'amb', 'ann', 'bwt', 'pac', 'sa', 'rbwt', 'rpac', 'rsa', \
            'stidx', 'sthash'],
        folderName="reference"):
        """
        suffix here doesn't include ".".
        
        deduce needBWARefIndexJob, needSAMtoolsFastaIndexJob, needPicardFastaDictJob,
            needStampyRefIndexJob from missing suffixes.
        argument folderName
        argument "addPicardDictFile" to offer user option to exclude this file 
            (i.e. in registerBlastNucleotideDatabaseFile).
        
        .dict is output by picard, also required for GATK.
            fai is via "samtools faidx" (index reference). also required for GATK
            amb', 'ann', 'bwt', 'pac', 'sa', 'rbwt', 'rpac', 'rsa' are all bwa index.
            stidx is stampy index.
            sthash is stampy hash.
        stidx (stampy index) and sthash (stampy hash)
        2011-11-11
            if needAffiliatedFiles,
                all other files, with suffix in affiliateFilenameSuffixLs,
                 will be registered (symlinked or copied) as well.
        """
        returnData = PassingData(refFastaFList = [],
            needBWARefIndexJob=False,
            needSAMtoolsFastaIndexJob=False,
            needPicardFastaDictJob=False,
            needStampyRefIndexJob=False,
            needBlastMakeDBJob=False,
            refPicardFastaDictF=None, refSAMtoolsFastaIndexF=None)
        missingSuffixSet = set()
        
        if registerAffiliateFiles:
            ref_fasta_file = self.registerOneInputFile(input_path=refFastaFname, folderName=folderName)
            returnData.refFastaFList.append(ref_fasta_file)
            # If it's not needed, assume the index is done and all relevant files
            #  are in absolute path. and no replica transfer
            
            #add extra affiliated files
            suffix2PathToFileLs = {}
            if addPicardDictFile:
                picardDictSuffix = 'dict'
                pathToFile = '%s.%s'%(os.path.splitext(refFastaFname)[0], \
                    picardDictSuffix)
                #remove ".fasta" from refFastaFname
                if checkAffiliateFileExistence and not os.path.isfile(pathToFile):
                    logging.warn(f"{pathToFile} don't exist or not a file. "
                        "Skip registration.")
                    missingSuffixSet.add(picardDictSuffix)
                    #suffix2PathToFileLs.append(pathToFile)
                else:
                    suffix2PathToFileLs[picardDictSuffix] = pathToFile
            for suffix in affiliateFilenameSuffixLs:
                pathToFile = '%s.%s'%(refFastaFname, suffix)
                if checkAffiliateFileExistence and not os.path.isfile(pathToFile):
                    logging.warn(f"{pathToFile} don't exist or not a file. "
                        "Skip registration.")
                    missingSuffixSet.add(suffix)
                    continue
                suffix2PathToFileLs[suffix]= pathToFile
            for suffix, pathToFile in suffix2PathToFileLs.items():
                if checkAffiliateFileExistence and not os.path.isfile(pathToFile):
                    logging.warn(f"{pathToFile} don't exist or not a file. "
                        "Skip registration.")
                    continue
                affiliate_file = self.registerOneInputFile(input_path=pathToFile,
                    folderName=folderName)
                returnData.refFastaFList.append(affiliate_file)
                
                if suffix=='dict':
                    returnData.refPicardFastaDictF = affiliate_file
                elif suffix=='fai':
                    returnData.refSAMtoolsFastaIndexF = affiliate_file
        else:
            ref_fasta_file = self.registerOneInputFile(input_path=refFastaFname, folderName=folderName)
            returnData.refFastaFList.append(ref_fasta_file)
        if 'bwt' in missingSuffixSet or 'pac' in missingSuffixSet:
            returnData.needBWARefIndexJob = True
        if 'fai' in missingSuffixSet:
            returnData.needSAMtoolsFastaIndexJob = True
            returnData.needPicardFastaDictJob = True
        if 'stidx' in missingSuffixSet or 'sthash' in missingSuffixSet:
            returnData.needStampyRefIndexJob = True
        if 'dict' in missingSuffixSet:
            returnData.needPicardFastaDictJob = True
        if 'nin' in missingSuffixSet or 'nhr' in missingSuffixSet or \
            'nsq' in missingSuffixSet:
            returnData.needBlastMakeDBJob = True
        return returnData

    def registerExecutables(self):
        """
        """
        ParentClass.registerExecutables(self)
        #for OutputVCFSiteStat.py
        self.registerOneExecutable(
            path=os.path.join(self.pymodulePath, "mapper/extractor/tabixRetrieve.sh"),
            name='tabixRetrieve', clusterSizeMultiplier=1)

        # from palos/polymorphism/FindNewRefCoordinatesGivenVCFFolder.py
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, \
            "polymorphism/mapper/LiftOverVCFBasedOnCoordinateMap.py"), \
            name='LiftOverVCFBasedOnCoordinateMap', clusterSizeMultiplier=1)

        self.registerOneExecutable(path=os.path.join(self.pymodulePath, \
            "polymorphism/qc/"+\
            "CalculateLociAndGenomeCoveredAtEachSwitchFrequencyThreshold.py"),
            name='CalculateLociAndGenomeCoveredAtEachSwitchFrequencyThreshold',
            clusterSizeMultiplier=0.01)

        self.registerOneExecutable(path=os.path.join(self.pymodulePath, \
                "mapper/extractor/ExtractFlankingSequenceForVCFLoci.py"), \
            name='ExtractFlankingSequenceForVCFLoci', clusterSizeMultiplier=2)

        self.registerOneExecutable(path=os.path.join(self.pymodulePath, \
            "polymorphism/mapper/FindSNPPositionOnNewRefFromFlankingBlastOutput.py"), \
            name='FindSNPPositionOnNewRefFromFlankingBlastOutput',
            clusterSizeMultiplier=2)

        self.registerOneExecutable(path=os.path.join(self.pymodulePath, \
            "polymorphism/mapper/FindSNPPositionOnNewRefFromFlankingBWAOutput.py"), \
            name='FindSNPPositionOnNewRefFromFlankingBWAOutput',
            clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, \
            'db/export/OutputGenomeAnnotation.py'), \
            name='OutputGenomeAnnotation', clusterSizeMultiplier=0.01)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, \
                'statistics/GenomeMovingAverageStatistics.py'), \
            name='GenomeMovingAverageStatistics', clusterSizeMultiplier=0.1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, \
                'mapper/computer/OutputVCFSiteGap.py'), \
            name='OutputVCFSiteGap', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, \
                'plot/PlotLD.py'), \
            name='PlotLD', clusterSizeMultiplier=0)
        
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            'mapper/converter/ConvertBjarniSNPFormat2Yu.py'), \
            name='ConvertBjarniSNPFormat2Yu', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            'mapper/converter/ConvertVCF2BjarniFormat.py'), \
            name='ConvertVCF2BjarniFormat', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            'mapper/converter/ConvertYuSNPFormat2Bjarni.py'), \
            name='ConvertYuSNPFormat2Bjarni', clusterSizeMultiplier=1)

        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            'mapper/converter/ConvertYuSNPFormat2EigenStrat.py'), \
            name='ConvertYuSNPFormat2EigenStrat', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            'mapper/converter/ConvertYuSNPFormat2TPED_TFAM.py'), \
            name='ConvertYuSNPFormat2TPED_TFAM', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            'mapper/CalculatePairwiseDistanceOutOfSNPXStrainMatrix.py'),
            name='CalculatePairwiseDistanceOutOfSNPXStrainMatrix',
            clusterSizeMultiplier=0.5)
        #use samtools to extract consensus from bam files
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            'mapper/alignment/ExtractConsensusSequenceFromAlignment.sh'),
            name='ExtractConsensusSequenceFromAlignment',
            clusterSizeMultiplier=0.5)
        #wrapper around psmc's splitfa,
        #  a program that splits fasta files
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
                "mapper/splitter/splitfa.sh"), \
            name='splitfa', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
                'mapper/modifier/AppendExtraPedigreeIndividualsToTPED.py'),
            name='AppendExtraPedigreeIndividualsToTPED', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
                'mapper/converter/ConvertMSOutput2FASTQ.py'),
            name='ConvertMSOutput2FASTQ', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
                'mapper/extractor/SelectChromosomeSequences.py'),
            name='SelectChromosomeSequences', clusterSizeMultiplier=0.5)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
                'plot/PlotGenomeWideData.py'),
            name='PlotGenomeWideData', clusterSizeMultiplier=1)

if __name__ == '__main__':
    from argparse import ArgumentParser
    ap = ArgumentParser()
    ap.add_argument("-i", "--input_path", type=str, required=True,
        help="the path to the input folder or file.")
    ap.add_argument("--inputSuffixList", type=str,
        help='Coma-separated list of input file suffices. Used to exclude input files.'
        'If None, no exclusion. The dot is part of the suffix, .tsv not tsv.'
        'Common zip suffices (.gz, .bz2, .zip, .bz) will be ignored in obtaining the suffix.')
    ap.add_argument("-F", "--pegasusFolderName", type=str,
        help='The path relative to the workflow running root. '
        'This folder will contain pegasus input & output. '
        'It will be created during the pegasus staging process. '
        'It is useful to separate multiple sub-workflows. '
        'If empty or None, everything is in the pegasus root.')
    
    ap.add_argument("-o", "--output_path", type=str, required=True,
        help="The path to the output file that will contain the Pegasus DAG.")
    
    ap.add_argument("--home_path", type=str,
        help="Path to your home folder. Default is ~.")
    ap.add_argument("--javaPath", type=str, default='bin/java',
        help="Path to java. "
        "If relative path, home folder is inserted in the front.")
    ap.add_argument("--pymodulePath", type=str, default="src/pymodule",
        help="Path to the pymodule code folder. "
        "If relative path, home folder is inserted in the front.")
    ap.add_argument("--thisModulePath", type=str,
        help="Path to your own code repo"
        "If relative path, home folder is inserted in the front.")
    ap.add_argument("--plinkPath", type=str, default='bin/plink',
        help="Path to plink, "
        "http://pngu.mgh.harvard.edu/~purcell/plink/index.shtml "
        "If relative path, home folder is inserted in the front.")
    
    ap.add_argument("--tmpDir", type=str, default='/tmp/',
        help='Default: %(default)s. '
        'A local folder for some jobs (MarkDup) to store temp data. '
        '/tmp/ can be too small for high-coverage sequencing.')
    ap.add_argument("--max_walltime", type=int, default=4320,
        help='Default: %(default)s. '
        'Maximum wall time for any job, in minutes. 4320=3 days. '
        'Used in addGenericJob(). Most clusters have upper limit for runtime.')
    ap.add_argument("--jvmVirtualByPhysicalMemoryRatio", type=float,
        default=1.2,
        help='Default: %(default)s. '
        'If a job virtual memory (~1.2X of JVM resident memory) exceeds request, '
        "it will be killed on some clusters. "
        "This will make sure your job has enough memory.")
    
    ap.add_argument("-l", "--site_handler", type=str, required=True,
        help="The name of the computing site where the jobs run and executables are stored. "
        "Check your Pegasus configuration in submit.sh.")
    ap.add_argument("-j", "--input_site_handler", type=str,
        help="It is the name of the site that has all the input files."
        "Possible values can be 'local' or same as site_handler."
        "If not given, it is asssumed to be the same as site_handler and "
        "the input files will be symlinked into the running folder."
        "If input_site_handler=local, the input files will be transferred "
        "to the computing site by pegasus-transfer.")
    ap.add_argument("-C", "--cluster_size", type=int, default=30,
        help="Default: %(default)s. "
        "This number decides how many of pegasus jobs should be clustered into one job. "
        "Good if your workflow contains many quick jobs. "
        "It will reduce Pegasus monitor I/O.")
    
    ap.add_argument("--debug", action='store_true',
        help='Toggle debug mode.')
    ap.add_argument("--report", action='store_true',
        help="Toggle verbose mode. Default: %(default)s.")
    ap.add_argument("--needSSHDBTunnel", action='store_true',
        help="If all DB-interacting jobs need a ssh tunnel to "
        "access a database that is inaccessible to computing nodes.")
    args = ap.parse_args()
    instance = AbstractBioinfoWorkflow(
        input_path=args.input_path,
        inputSuffixList=args.inputSuffixList, 
        pegasusFolderName=args.pegasusFolderName,
        output_path=args.output_path,

        home_path=args.home_path,
        javaPath=args.javaPath,
        pymodulePath=args.pymodulePath,
        thisModulePath=args.thisModulePath,
        
        tmpDir=args.tmpDir,
        max_walltime=args.max_walltime,
        jvmVirtualByPhysicalMemoryRatio=args.jvmVirtualByPhysicalMemoryRatio,
        
        site_handler=args.site_handler, 
        input_site_handler=args.input_site_handler,
        cluster_size=args.cluster_size,

        needSSHDBTunnel=args.needSSHDBTunnel,
        debug=args.debug,
        report=args.report)
    instance.setup_run()
    instance.end_run()



#!/usr/bin/env python3
"""
2012.5.23
    a common class for other pymodule workflows
"""
import sys, os
from palos import PassingData
from pegaflow.Workflow import Workflow
from pegaflow.api import File, Job

#path to the source code directory
src_dir = os.path.dirname(os.path.abspath(__file__))

class AbstractWorkflow(Workflow):
    __doc__ = __doc__
    db_option_dict = {
        ('drivername', 1,):['postgresql', 'v', 1, \
            'which type of database? mysql or postgresql', ],\
        ('hostname', 1, ): ['localhost', 'z', 1, 'hostname of the db server', ],\
        ('dbname', 1, ): ['', 'd', 1, 'database name', ],\
        ('schema', 0, ): ['public', 'k', 1, 'database schema name', ],\
        ('db_user', 1, ): [None, 'u', 1, 'database username', ],\
        ('db_passwd', 1, ): [None, 'p', 1, 'database password', ],\
        ('port', 0, ):[None, '', 1, 'database port number'],\
        ('commit', 0, ):[None, '', 0, 
            'commit database transaction if there is db transaction'],
        ("data_dir", 0, ): ["", 't', 1, \
            'the base directory where all db-affiliated files are stored. '
            'If not given, use the default stored in db.'],\
        ("local_data_dir", 0, ): ["", 'D', 1, 
            'A folder contains the same files as data_dir but accessible locally. '
            'If not given, use the default stored in db (db.data_dir). '
            'This argument is used to find all input files available. '
            'Set it differently from data_dir only if you are '
            'generating a workflow on a computer that '
            'has a different path to data_dir than your running site.'],
        }
    option_default_dict = {
        ("home_path", 1, ): [os.path.expanduser("~"), 'e', 1, \
            'path to the home directory on the working nodes'],\
        ("pymodulePath", 1, ): ["src/pymodule", '', 1, 'path to the pymodule folder'],\
        ("thisModulePath", 1, ): ["", '', 1, 
            'path of the module that owns this program. '
            'used to add executables from this module.'],\
        ("javaPath", 1, ): ["bin/java", 'J', 1, 'the path to java binary'],\
        ("site_handler", 1, ): ["condor", 'l', 1, 
            'which computing site to run the jobs. check your Pegasus setup.'],\
        ("input_site_handler", 0, ): ["None", 'j', 1, 
            'the site that stores all the input files: local, condor, hoffman2. '
            'If site_handler is condor, this must be condor and files will be symlinked. '
            'If site_handler is hoffman2, input_site_handler=local induces file transfer '
            'and input_site_handler=hoffman2 induces symlink. '
            'Default is the same as the site_handler.'],\
        ('cluster_size', 1, int):[30, 'C', 1, 
            'For short jobs that will be clustered, how many of them should be clustered int one'],\
        ('pegasusFolderName', 0, ): ['folder', 'F', 1, \
            'The folder relative to pegasus scratch working dir to contain input & output. '
            'It will be created during the pegasus staging process. '
            'It is useful to separate multiple workflows. '
            'If empty, everything is in the pegasus root.', ],\
        ('inputSuffixList', 0, ): [None, '', 1, \
            'coma-separated list of input file suffices. If None, any suffix. '
            'Suffix include the dot, (i.e. .tsv). '
            'Typical zip suffices are excluded (.gz, .bz2, .zip, .bz).'],\
        ('output_path', 1, ): [None, 'o', 1, 'xml workflow output file'],\
        ("tmpDir", 1, ): ["/tmp/", '', 1, \
            'for MarkDuplicates.jar, etc., default is /tmp/ but sometimes it is too small'],\
        ('max_walltime', 1, int):[4320, '', 1, \
            'maximum wall time any job could have, in minutes. 20160=2 weeks.\n'
            'used in addGenericJob().'],\
        ('jvmVirtualByPhysicalMemoryRatio', 1, float):[1.2, '', 1, 
            "if a job's virtual memory (usually 1.2X of JVM resident memory) exceeds request, "
            "it will be killed on hoffman2. Hence this argument"],\
        ('debug', 0, int):[0, 'b', 0, 'toggle debug mode'],\
        ('needSSHDBTunnel', 0, int):[0, 'H', 0, \
            'DB-interacting jobs need a ssh tunnel (running on cluster behind firewall).'],\
        ('report', 0, int):[0, 'r', 0, 'toggle report, more verbose stdout/stderr.']
        }
        #("pegasusCleanupPath", 1, ): ["bin/pegasus/bin/pegasus-cleanup", '', 1, \
        #    'path to pegasus-cleanup executable, it will be registered to run in local '
        #    'universe of condor pool (rather than the vanilla universe)'],\
        #("pegasusTransferPath", 1, ): ["bin/pegasus/bin/pegasus-transfer", '', 1, \
        #    'path to pegasus-transfer executable, it will be registered to run in local '
        #     'universe of condor pool (rather than the vanilla universe)'],\
        #('bamListFname', 1, ): ['/tmp/bamFileList.txt', 'L', 1, 
        #     'The file contains path to each bam file, one file per line.'],\

    pathToInsertHomePathList = ['javaPath', 'pymodulePath', 'plinkPath', "thisModulePath"]

    def __init__(self,
        input_path=None,
        inputSuffixList=None, 
        pegasusFolderName='folder',
        output_path=None,

        tmpDir='/tmp/', max_walltime=4320,

        home_path=None,
        pymodulePath="src/pymodule", thisModulePath=None,
        javaPath=None, 
        jvmVirtualByPhysicalMemoryRatio=1.2,
        
        site_handler='condor', input_site_handler='condor',
        cluster_size=30,
        needSSHDBTunnel=False, commit=False,
        debug=False, report=False):
        """
        """
        # Set before ParentClass.__init__()
        if not home_path:
            home_path = os.path.expanduser("~")
        self.pymodulePath = pymodulePath
        self.thisModulePath = thisModulePath
        Workflow.__init__(self,
            input_path=input_path,
            inputSuffixList=inputSuffixList,
            pegasusFolderName=pegasusFolderName,
            output_path=output_path,

            tmpDir=tmpDir, max_walltime=max_walltime, 
            
            home_path=home_path,
            javaPath=javaPath,
            jvmVirtualByPhysicalMemoryRatio=jvmVirtualByPhysicalMemoryRatio,
            
            site_handler=site_handler,
            input_site_handler=input_site_handler,
            cluster_size=cluster_size,

            needSSHDBTunnel=needSSHDBTunnel, commit=commit,
            debug=debug, report=report)

    def constructJobDataFromJob(self, job=None):
        """
        added vcfFile and tbi_F in structure
        """
        if hasattr(job, 'output') and job.output.name.find('.vcf')>=0:
            vcfFile = job.output
            tbi_F = getattr(job, 'tbi_F', None)
        else:
            vcfFile = None
            tbi_F = None
        return PassingData(job=job, jobLs=[job], file=job.output,
            fileLs=job.outputLs, vcfFile=vcfFile, tbi_F=tbi_F)

    def registerExecutables(self):
        """
        """
        Workflow.registerExecutables(self)
        
        if os.path.isfile("/usr/bin/convert"):
            #convert, an image swissknife program, part of imagemagick
            self.registerOneExecutable(path="/usr/bin/convert",
                name='convertImage', clusterSizeMultiplier=1)
        self.SelectLineBlockFromFile = self.registerOneExecutable(
            path=os.path.join(self.pymodulePath, "mapper/extractor/SelectLineBlockFromFile.py"), 
            clusterSizeMultiplier=1)
        self.registerOneExecutable(
            path=os.path.join(self.pymodulePath, "palos/plot/AbstractPlot.py"),
            clusterSizeMultiplier=1)
        self.registerOneExecutable(
            path=os.path.join(self.pymodulePath, "plot/PlotYAsBar.py"),
            clusterSizeMultiplier=1)
        self.DrawHistogram = self.registerOneExecutable(
            path=os.path.join(self.pymodulePath, "plot/DrawHistogram.py"),
            clusterSizeMultiplier=1)

        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            "plot/DrawMatrix.py"), clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            "plot/Draw2DHistogramOfMatrix.py"), clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            "statistics/SampleRows.py"), 
            clusterSizeMultiplier=1)
        #2013.2.11 all reducers
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            "statistics/EstimateOutliersIn2DData.py"), \
            clusterSizeMultiplier=0)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            'reducer/MergeSameHeaderTablesIntoOne.py'), \
            name='mergeSameHeaderTablesIntoOne',
            clusterSizeMultiplier=0)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            'reducer/MergeSameHeaderTablesIntoOne.py'), \
            clusterSizeMultiplier=0)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            'reducer/ReduceMatrixByAverageColumnsWithSameKey.py'), \
            clusterSizeMultiplier=0)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            'reducer/ReduceMatrixByChosenColumn.py'), \
            clusterSizeMultiplier=0)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            'reducer/ReduceMatrixByMergeColumnsWithSameKey.py'), \
            clusterSizeMultiplier=0)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            'reducer/ReduceMatrixBySumSameKeyColsAndThenDivide.py'),
            clusterSizeMultiplier=0)
        """
        # 2013.05.20 DISABLE this
        if self.site_handler=='hcondor' and self.input_site_handler=='hcondor':
            #2013.04.19 to make pegasus cleanup run on local universe of condor pool
            # only enable this on hcondor because
            #	1) its filesystem is very slow and these cleanup & transfer
            #       jobs take forever.
            #	2) workers in vanilla universe expire after certain time.
            #	3) it does not run on ycondor local universe somehow.
            #    pegasus keeps submitting but no condor jobs in the queue.
            # this works because in most of my cases, vanilla universe and
            #    local universe share the same underlying filesystem.
            cleanupExecutable = self.registerOneExecutable(
                path=self.pegasusCleanupPath, name='cleanup',
                clusterSizeMultiplier=0, noVersion=True)
            condorUniverseProfile = Profile(Namespace.CONDOR, key="universe",
                value="local")
            if cleanupExecutable.hasProfile(condorUniverseProfile):
                cleanupExecutable.removeProfile(condorUniverseProfile)
            cleanupExecutable.addProfile(condorUniverseProfile)

            transferExecutable = self.registerOneExecutable(
                path=self.pegasusTransferPath, name='transfer',
                clusterSizeMultiplier=0, noVersion=True)
            condorUniverseProfile = Profile(Namespace.CONDOR, key="universe",
                value="local")
            if transferExecutable.hasProfile(condorUniverseProfile):
                transferExecutable.removeProfile(condorUniverseProfile)
            transferExecutable.addProfile(condorUniverseProfile)
        """

    def addStatMergeJob(self, statMergeProgram=None, outputF=None,
        extraArguments=None, extraArgumentList=None,
        parentJobLs=None, extraOutputLs=None,
        extraDependentInputLs=None, transferOutput=True,
        key2ObjectForJob=None, job_max_memory=1000, **keywords):
        """
        moved from CalculateVCFStatPipeline.py
        """
        if extraDependentInputLs is None:
            extraDependentInputLs = []
        if extraArgumentList is None:
            extraArgumentList = []
        if key2ObjectForJob is None:
            key2ObjectForJob = {}
        if extraArguments:
            extraArgumentList.append(extraArguments)
        job= self.addGenericJob(executable=statMergeProgram,
            inputFile=None, inputArgumentOption='-i',
            outputFile=outputF, outputArgumentOption='-o', 
            extraArgumentList=extraArgumentList,
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs,
            extraOutputLs=extraOutputLs,
            transferOutput=transferOutput,
            key2ObjectForJob=key2ObjectForJob,
            job_max_memory=job_max_memory, **keywords)
        return job

    def addConvertImageJob(self, inputFile=None, inputArgumentOption=None,
        outputFile=None, outputArgumentOption=None, density=None,
        resizeDimension=None,
        frontArgumentList=None,
        extraArguments=None, extraArgumentList=None,
        parentJobLs=None,
        extraDependentInputLs=None, extraOutputLs=None,
        transferOutput=False,
        job_max_memory=200,
        key2ObjectForJob=None, **keywords):
        """
        2013.2.7 use imagemagick's convert to convert images. examples:
            plotOutputFile = File('%s.eps'%(plotOutputFnamePrefix))
            plotPNGOutputFile = File('%s.png'%(plotOutputFnamePrefix))
            #change dpi to 300
            self.addConvertImageJob(inputFile=plotOutputFile, \
                outputFile=plotPNGOutputFile, density=300, \
                resizeDimension=None, \
                parentJobLs=[psmc_plotJob], extraDependentInputLs=None, 
                extraOutputLs=None, transferOutput=True, \
                extraArguments=None, frontArgumentList=None, job_max_memory=500)

            #resize by demanding the width = 1800, height will scale accordingly
            self.addConvertImageJob(inputFile=plotOutputFile, \
                outputFile=plotPNGOutputFile, density=None, \
                resizeDimension=1800, \
                parentJobLs=[psmc_plotJob], extraDependentInputLs=None, \
                extraOutputLs=None, transferOutput=True, \
                extraArguments=None, frontArgumentList=None, job_max_memory=500)
            #resize by demanding the dimension=1800X900
            self.addConvertImageJob(inputFile=plotOutputFile, \
                outputFile=plotPNGOutputFile, density=None, \
                resizeDimension='1800X900', \
                parentJobLs=[psmc_plotJob], extraDependentInputLs=None, \
                extraOutputLs=None, transferOutput=True, \
                extraArguments=None, frontArgumentList=None, job_max_memory=500)
        """
        if frontArgumentList is None:
            frontArgumentList = []
        if extraOutputLs is None:
            extraOutputLs = []
        if density is not None:
            frontArgumentList.append("-density %s"%(density))
        if resizeDimension is not None:
            frontArgumentList.append("-resize %s"%(resizeDimension))
        #do not pass the inputFileList to addGenericJob() because db arguments
        #  need to be added before them.
        job = self.addGenericJob(executable=self.convertImage,
            inputFile=inputFile,
            inputArgumentOption=inputArgumentOption,
            outputFile=outputFile,
            outputArgumentOption=outputArgumentOption,
            inputFileList=None, 
            frontArgumentList=frontArgumentList,
            extraArguments=extraArguments, 
            extraArgumentList=extraArgumentList,
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs,
            extraOutputLs=extraOutputLs,
            transferOutput=transferOutput,
            job_max_memory=job_max_memory,
            key2ObjectForJob=key2ObjectForJob,
            **keywords)
        return job
    
    def addCalculateDepthMeanMedianModeJob(self, executable=None,
        inputFile=None, outputFile=None, alignmentID=None,
        fractionToSample=0.001,
        whichColumn=None, maxNumberOfSamplings=1E7,
        inputStatName=None,
        extraArguments=None,
        parentJobLs=None, job_max_memory = 500,
        transferOutput=False, **keywords):
        """
        2012.6.15 turn maxNumberOfSamplings into integer when passing it to the job
        input is output of samtoolsDepth.
        """
        extraArgumentList = []
        if alignmentID is not None:
            extraArgumentList.append("--alignmentID %s"%(alignmentID))
        if fractionToSample is not None:
            extraArgumentList.append("--fractionToSample %s"%(fractionToSample))
        if whichColumn is not None:
            extraArgumentList.append("--whichColumn %s"%(whichColumn))
        if maxNumberOfSamplings is not None:
            extraArgumentList.append('--maxNumberOfSamplings %d'%(maxNumberOfSamplings))
        if inputStatName is not None:
            extraArgumentList.append("--inputStatName %s"%(inputStatName))
        if extraArguments:
            extraArgumentList.append(extraArguments)
        job= self.addGenericJob(executable=executable,
            inputFile=inputFile,
            inputArgumentOption='-i',
            outputFile=outputFile,
            outputArgumentOption='-o',
            extraArgumentList=extraArgumentList,
            parentJobLs=parentJobLs,
            transferOutput=transferOutput,
            job_max_memory=job_max_memory, **keywords)
        return job

    def addDBGenomeArgumentsToOneJob(self, job:Job=None, objectWithDBArguments=None):
        """
        2013.07.31 similar to addDBArgumentsToOneJob() but for genome db
        """
        if objectWithDBArguments is None:
            objectWithDBArguments = self
        if self.drivername=='mysql':
            genome_dbname = 'genome'
        else:
            genome_dbname = self.dbname
        job.add_args(
            "--genome_drivername", objectWithDBArguments.drivername,
            "--genome_hostname", objectWithDBArguments.hostname, \
            "--genome_dbname", genome_dbname, \
            "--genome_db_user", objectWithDBArguments.db_user, \
            "--genome_db_passwd %s"%objectWithDBArguments.db_passwd)

        job.add_args("--genome_schema genome")

        if getattr(objectWithDBArguments, 'port', None):
            job.add_args("--genome_port=%s"%(objectWithDBArguments.port))
        return job


    def addGzipSubWorkflow(self, inputData=None, transferOutput=True,\
        outputDirPrefix="", topOutputDirJob=None, **keywords):
        """
        2012.8.2 bugfix.
        2012.7.19
        """
        sys.stderr.write("Adding gzip jobs for %s input job data ... "%(
            len(inputData.jobDataLs)))
        returnData = PassingData(topOutputDirJob=None)
        returnData.jobDataLs = []
        if inputData:
            if len(inputData.jobDataLs)>0:
                if topOutputDirJob is None:
                    topOutputDir = "%sGzip"%(outputDirPrefix)
                    topOutputDirJob = self.addMkDirJob(outputDir=topOutputDir)

                returnData.topOutputDirJob = topOutputDirJob
                for jobData in inputData.jobDataLs:
                    for inputF in set(jobData.fileLs):
                        #2013.08.16 do not work on same file
                        inputFBaseName = os.path.basename(inputF.name)
                        outputF = File(os.path.join(topOutputDirJob.output, \
                            '%s.gz'%(inputFBaseName)))
                        key2ObjectForJob = {}
                        extraArgumentList = []
                        #make sure set inputArgumentOption&outputArgumentOption to None, \
                        # otherwise addGenericJob will add "-i" and "-o" in front of it
                        job= self.addGenericJob(executable=self.gzip,
                            inputArgumentOption=None,
                            inputFile=inputF,
                            outputArgumentOption=None,
                            outputFile=outputF,
                            extraArgumentList=extraArgumentList,
                            parentJobLs=[topOutputDirJob]+jobData.jobLs,
                            extraDependentInputLs=None,
                            extraOutputLs=None,
                            transferOutput=transferOutput,
                            key2ObjectForJob=key2ObjectForJob,
                            job_max_memory=200, **keywords)
                        returnData.jobDataLs.append(PassingData(jobLs=[job],
                            vcfFile=None, file=outputF, fileLs=[outputF]))
        sys.stderr.write("no_of_jobs = %s.\n"%(self.no_of_jobs))
        return returnData

    def addAbstractMapperLikeJob(self, executable=None, \
        inputVCF=None, inputF=None, outputF=None,
        extraArguments=None, extraArgumentList=None,
        extraOutputLs=None,\
        parentJobLs=None, transferOutput=True, job_max_memory=200,\
        extraDependentInputLs=None, \
        sshDBTunnel=None, **keywords):
        """
        2012.10.8 call addGenericJob() instead
        2012.7.19
            moved from AbstractNGSWorkflow to here.
            add argument inputF. inputVCF is not generic enough.
        2012.5.11
        """
        if inputF is None:	#2012.7.19
            inputF = inputVCF
        job= self.addGenericJob(executable=executable, inputFile=inputF,
            outputFile=outputF, \
            extraArguments=extraArguments,\
            extraArgumentList=extraArgumentList,
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs,
            extraOutputLs=extraOutputLs,\
            transferOutput=transferOutput, \
            sshDBTunnel=sshDBTunnel, job_max_memory=job_max_memory, **keywords)
        return job

    def addSelectLineBlockFromFileJob(self,
        executable=None,
        inputFile=None, outputFile=None,
        startLineNumber=None, stopLineNumber=None,
        extraArguments=None,
        parentJobLs=None,
        extraDependentInputLs=None,
        transferOutput=False,
        job_max_memory=200, **keywords):
        """
        """
        extraArgumentList = ['-s %s'%(startLineNumber), '-t %s'%(stopLineNumber)]
        if extraArguments:
            extraArgumentList.append(extraArguments)

        job= self.addGenericJob(executable=executable, inputFile=inputFile,
            outputFile=outputFile, \
            extraArgumentList=extraArgumentList,
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs, \
            extraOutputLs=None,\
            transferOutput=transferOutput, \
            job_max_memory=job_max_memory, \
            **keywords)
        return job

    def addPlotLDJob(self, executable=None,
        inputFile=None, inputFileList=None, \
        outputFile=None, outputFnamePrefix=None,
        whichColumn=None, whichColumnHeader=None, whichColumnPlotLabel=None,
        title=None, logY=None, valueForNonPositiveYValue=-1,
        missingDataNotation='-nan',
        xColumnHeader=None,
        xColumnPlotLabel=None,
        chrLengthColumnHeader=None, chrColumnHeader=None,
        minChrLength=1000000,
        pos2ColumnHeader=None,
        minNoOfTotal=100, maxNoOfTotal=None,
        figureDPI=300, formatString='.', ylim_type=None,
        samplingRate=1, 
        need_svg=False, logCount=False,
        minDist=None, maxDist=None, movingAverageType=2,
        extraArguments=None,
        extraArgumentList=None,
        parentJobLs=None,
        extraDependentInputLs=None,
        transferOutput=True,
        job_max_memory=200, **keywords):
        """
('outputFname', 1, ): [None, 'o', 1, 'output file for the figure.'],\
('minNoOfTotal', 1, int): [100, 'i', 1, 
    'minimum no of total variants (denominator of inconsistent rate)'],\
('title', 1, ): [None, 't', 1, 'title for the figure.'],\
('figureDPI', 1, int): [200, 'f', 1, 'dpi for the output figures (png)'],\
('formatString', 1, ): ['.', 'a', 1, 'formatString passed to matplotlib plot'],\
('ylim_type', 1, int): [1, 'y', 1, 'y-axis limit type, 1: 0 to max. 2: min to max'],\
('samplingRate', 1, float): [0.001, 's', 1, 'how often you include the data'],\
('debug', 0, int):[0, 'b', 0, 'toggle debug mode'],\
('report', 0, int):[0, 'r', 0, 'toggle report, more verbose stdout/stderr.'],\
('whichColumn', 0, int): [3, 'w', 1, 
    'data from this column (index starting from 0) is plotted as y-axis value'],\
('whichColumnHeader', 0, ): ["", 'W', 1, 
    'column label (in the header) for the data to be plotted as y-axis value, substitute whichColumn'],\
('logWhichColumn', 0, int): [0, 'g', 0, 'whether to take -log of the whichColumn'],\
('whichColumnPlotLabel', 1, ): ['#SNPs in 100kb window', 'D', 1, \
    'plot label for data of the whichColumn', ],\
('chrLengthColumnHeader', 1, ): ['chrLength', 'c', 1, 'label of the chromosome length column', ],\
('chrColumnHeader', 1, ): ['CHR', 'C', 1, 'label of the chromosome column', ],\
('minChrLength', 1, int): [1000000, 'm', 1, \
    'minimum chromosome length for one chromosome to be included', ],\
('pos1ColumnLabel', 1, ): ['POS1', 'l', 1, 'label of the 1st position column', ],\
('pos2ColumnLabel', 1, ): ['POS2', 'p', 1, 'label of the 2nd position column', ],\
('posColumnPlotLabel', 1, ): ['distance', 'x', 1, 'x-axis label in  plot', ],\

        """
        if extraArguments is None:
            extraArguments = ""
        if extraArgumentList is None:
            extraArgumentList = []
        if logCount:
            extraArguments += " --logCount "
        if minChrLength is not None:
            extraArguments += " --minChrLength %s "%(minChrLength)
        if chrLengthColumnHeader:
            extraArgumentList.append("--chrLengthColumnHeader %s"%\
                (chrLengthColumnHeader))
        if chrColumnHeader:
            extraArgumentList.append("--chrColumnHeader %s"%(chrColumnHeader))
        if pos2ColumnHeader:
            extraArgumentList.append(' --pos2ColumnHeader %s '%(pos2ColumnHeader))
        if minDist:
            extraArgumentList.append('--minDist %s'%(minDist))
        if maxDist:
            extraArgumentList.append('--maxDist %s'%(maxDist))
        if maxNoOfTotal:
            extraArgumentList.append("--maxNoOfTotal %s"%(maxNoOfTotal))
        if movingAverageType:
            extraArgumentList.append("--movingAverageType %s"%(movingAverageType))

        return self.addAbstractPlotJob(executable=executable,
            inputFileList=inputFileList,
            inputFile=inputFile,
            outputFile=outputFile, outputFnamePrefix=outputFnamePrefix,
            xColumnHeader=xColumnHeader, xColumnPlotLabel=xColumnPlotLabel,
            whichColumn=whichColumn,
            whichColumnHeader=whichColumnHeader,
            whichColumnPlotLabel=whichColumnPlotLabel,
            logY=logY, valueForNonPositiveYValue=valueForNonPositiveYValue,
            missingDataNotation=missingDataNotation,
            title=title,
            minNoOfTotal=minNoOfTotal,
            figureDPI=figureDPI, formatString=formatString,
            ylim_type=ylim_type, 
            samplingRate=samplingRate, need_svg=need_svg,
            extraArguments=extraArguments,
            extraArgumentList=extraArgumentList,
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs,
            transferOutput=transferOutput,
            job_max_memory=job_max_memory,
            **keywords)

    def addPlotVCFtoolsStatJob(self,
        executable=None,
        inputFileList=None, 
        outputFnamePrefix=None,
        xColumnPlotLabel=None, xColumnHeader=None,
        whichColumn=None, whichColumnHeader=None, whichColumnPlotLabel=None, 
        need_svg=False, logY=0, valueForNonPositiveYValue=-1,
        chrLengthColumnHeader=None, 
        chrColumnHeader=None,
        minChrLength=1000000, minNoOfTotal=100,
        figureDPI=300, ylim_type=None, samplingRate=1,
        logCount=False,
        tax_id=60711, sequence_type_id=1, chrOrder=None,
        extraArguments=None,
        parentJobLs=None,
        extraDependentInputLs=None,
        transferOutput=True, job_max_memory=200,
        sshDBTunnel=False, **keywords):
        """
        Examples

    outputFnamePrefix = os.path.join(plotOutputDir, 'noOfMendelErrors_along_chromosome')
    self.addPlotVCFtoolsStatJob(executable=self.PlotVCFtoolsStat,
        inputFileList=[splitPlinkLMendelFileSNPIDIntoChrPositionJob.output],
        outputFnamePrefix=outputFnamePrefix,
        whichColumn=None, whichColumnHeader="N", whichColumnPlotLabel="noOfMendelErrors",
        need_svg=False,
        logY=0, valueForNonPositiveYValue=-1,
        xColumnPlotLabel="genomePosition",
        chrLengthColumnHeader=None, chrColumnHeader="Chromosome", \
        minChrLength=100000, xColumnHeader="Start", minNoOfTotal=50,\
        figureDPI=100, ylim_type=2, samplingRate=1,\
        tax_id=self.ref_genome_tax_id,
        sequence_type_id=self.ref_genome_sequence_type_id, chrOrder=1,\
        parentJobLs=[splitPlinkLMendelFileSNPIDIntoChrPositionJob, plotOutputDirJob], \
        extraDependentInputLs=None, \
        extraArguments=None, transferOutput=True, sshDBTunnel=self.needSSHDBTunnel)


        remove argument positiveLog, rename logWhichColumn to logY
        # whichColumnPlotLabel and xColumnPlotLabel should not contain spaces or ( or ).
        #  because they will disrupt shell commandline


('whichColumn', 0, int): [3, 'w', 1,
    'data from this column (index starting from 0) is plotted as y-axis value'],\
('whichColumnHeader', 0, ): ["", 'W', 1, 
    'column label (in the header) for the data to be plotted as y-axis value, substitute whichColumn'],\
('logY', 0, int): [0, '', 1, 'value 0: nothing; 1: log(), 2: -log(). replacing self.logWhichColumn.'],\
('need_svg', 0, ): [0, 'n', 0, 'whether need svg output', ],\
('whichColumnPlotLabel', 1, ): ['#SNPs in 100kb window', 'D', 1, 
    'plot label for data of the whichColumn', ],\
('xColumnPlotLabel', 1, ): ['position', 'x', 1, 'x-axis label (posColumn) in manhattan plot', ],\
('chrLengthColumnHeader', 1, ): ['chrLength', 'c', 1, 'label of the chromosome length column', ],\
('chrColumnHeader', 1, ): ['CHR', 'C', 1, 'label of the chromosome column', ],\
('minChrLength', 1, int): [1000000, 'm', 1, 'minimum chromosome length for one chromosome to be included', ],\
('xColumnHeader', 1, ): ['BIN_START', 'l', 1,
    'label of the position column, BIN_START for binned vcftools output. POS for others.', ],\
('outputFnamePrefix', 0, ): [None, 'O', 1, 'output filename prefix (optional).'],\

('minNoOfTotal', 1, int): [100, 'i', 1, 
    'minimum no of total variants (denominator of inconsistent rate)'],\
('title', 1, ): [None, 't', 1, 'title for the figure.'],\
('figureDPI', 1, int): [200, 'f', 1, 'dpi for the output figures (png)'],\
('formatString', 1, ): ['-', '', 1, 'formatString passed to matplotlib plot'],\
('ylim_type', 1, int): [1, 'y', 1, 'y-axis limit type, 1: 0 to max. 2: min to max'],\
('samplingRate', 1, float): [0.001, 's', 1, 'how often you include the data'],\
        """
        if extraDependentInputLs is None:
            extraDependentInputLs = []
        if inputFileList:
            extraDependentInputLs.extend(inputFileList)
        extraArgumentList = [
            '--xColumnHeader %s'%(xColumnHeader)
            ]
        extraOutputLs = [
            File('%s.png'%(outputFnamePrefix)),
            File('%s_hist.png'%(outputFnamePrefix))
            ]
        key2ObjectForJob = {}
        if ylim_type:
            extraArgumentList.append(f'--ylim_type {ylim_type}')
        if outputFnamePrefix:
            extraArgumentList.append(f'--outputFnamePrefix {outputFnamePrefix}')
        if minNoOfTotal:
            extraArgumentList.append(f'--minNoOfTotal {minNoOfTotal}')
        if figureDPI:
            extraArgumentList.append(f'--figureDPI {figureDPI}')
        if samplingRate:
            extraArgumentList.append(f'--samplingRate {samplingRate}')
        
        if need_svg:
            extraOutputLs.append(File('%s.svg'%(outputFnamePrefix)))
        if minChrLength is not None:
            extraArgumentList.append('--minChrLength %s'%(minChrLength))
        if whichColumnHeader:
            extraArgumentList.append("--whichColumnHeader %s"%(whichColumnHeader))
        if whichColumn:
            extraArgumentList.append("--whichColumn %s"%(whichColumn))
        if logY is not None:
            extraArgumentList.append('--logY %s'%(logY))
        if whichColumnPlotLabel:
            extraArgumentList.append("--whichColumnPlotLabel %s"%(whichColumnPlotLabel))
        if xColumnPlotLabel:
            extraArgumentList.append("--xColumnPlotLabel %s"%(xColumnPlotLabel))
        if chrLengthColumnHeader:
            extraArgumentList.append("--chrLengthColumnHeader %s"%\
                (chrLengthColumnHeader))
        if chrColumnHeader:
            extraArgumentList.append("--chrColumnHeader %s"%(chrColumnHeader))
        if logCount:
            extraArgumentList.append("--logCount")
        if valueForNonPositiveYValue:
            extraArgumentList.append("--valueForNonPositiveYValue %s"%\
                (valueForNonPositiveYValue))
        if sequence_type_id:
            extraArgumentList.append("--sequence_type_id %s"%(sequence_type_id))
        if tax_id:
            extraArgumentList.append("--tax_id %s"%(tax_id))
        if chrOrder is not None:
            extraArgumentList.append("--chrOrder %s"%(chrOrder))

        if extraArguments:
            extraArgumentList.append(extraArguments)
        job= self.addDBJob(
            executable=executable,
            inputFile=None,
            outputFile=None,
            inputFileList=inputFileList,
            extraArgumentList=extraArgumentList,
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs,
            extraOutputLs=extraOutputLs,
            transferOutput=transferOutput,
            key2ObjectForJob=key2ObjectForJob,
            job_max_memory=job_max_memory,
            sshDBTunnel=sshDBTunnel, objectWithDBArguments=self, **keywords)
        return job

    def addPlotGenomeWideDataJob(self, executable=None,
        inputFileList=None,
        inputFile=None,
        outputFnamePrefix=None, outputFile=None,
        xColumnPlotLabel=None, xColumnHeader=None,
        whichColumn=None, whichColumnHeader=None, whichColumnPlotLabel=None,
        logX=None, logY=None, valueForNonPositiveYValue=-1,
        xScaleLog=None, yScaleLog=None,
        missingDataNotation='NA',
        xtickInterval=None,
        drawCentromere=None, chrColumnHeader=None,
        minChrLength=100000, minNoOfTotal=100, maxNoOfTotal=None,
        figureDPI=300, formatString=".", ylim_type=1,
        samplingRate=1,
        logCount=False, need_svg=False,
        tax_id=60711, sequence_type_id=1, chrOrder=None,
        inputFileFormat=1, outputFileFormat=None,
        extraArguments=None,
        extraArgumentList=None,
        parentJobLs=None, extraDependentInputLs=None,
        transferOutput=True, job_max_memory=200, \
        objectWithDBGenomeArguments=None, sshDBTunnel=False, \
        **keywords):

        """
        Examples:
    outputFile = File(os.path.join(plotOutputDir, 'noOfMendelErrors_along_chromosome.png'))
    self.addPlotGenomeWideDataJob(inputFileList=None, \
        inputFile=splitPlinkLMendelFileSNPIDIntoChrPositionJob.output,\
        outputFile=outputFile,\
        whichColumn=None, whichColumnHeader="N",
        whichColumnPlotLabel="noOfMendelErrors", \
        logX=None, logY=None, valueForNonPositiveYValue=-1, \
        xScaleLog=None, yScaleLog=None,\
        missingDataNotation='NA',\
        xColumnPlotLabel="genomePosition", xColumnHeader="Start", \
        xtickInterval=20000000,\
        chrColumnHeader="Chromosome", \
        minChrLength=None, minNoOfTotal=None, maxNoOfTotal=None, \
        figureDPI=300, formatString=".", ylim_type=2, samplingRate=1, 
        logCount=False, need_svg=False,\
        tax_id=self.ref_genome_tax_id, 
        sequence_type_id=self.ref_genome_sequence_type_id, chrOrder=1,\
        inputFileFormat=1, outputFileFormat=None,\
        parentJobLs=[splitPlinkLMendelFileSNPIDIntoChrPositionJob, plotOutputDirJob], \
        extraDependentInputLs=None, \
        extraArguments=None, extraArgumentList=None, \
        transferOutput=True, job_max_memory=1000, sshDBTunnel=self.needSSHDBTunnel)

        """
        if extraArgumentList is None:
            extraArgumentList=[]
        if executable is None:
            executable = self.PlotGenomeWideData

        if objectWithDBGenomeArguments is None:
            objectWithDBGenomeArguments = self
        """
        #2013.07.31 replaced by addDBGenomeArgumentsToOneJob() in addGenericJob()
        extraArgumentList.extend(['--genome_drivername=%s'%self.drivername,\
                '--genome_hostname=%s'%self.hostname,\
                '--genome_dbname=%s'%(genome_dbname),\
                '--genome_schema=genome',\
                '--genome_db_user=%s'%(self.db_user),\
                '--genome_db_passwd=%s'%(self.db_passwd)])
        """
        if drawCentromere:
            extraArgumentList.append('--drawCentromere')
        if xtickInterval is not None:
            extraArgumentList.append("--xtickInterval %s"%(xtickInterval))
        if chrColumnHeader is not None:
            extraArgumentList.append('--chromosomeHeader %s'%(chrColumnHeader))
        if tax_id:
            extraArgumentList.append("--tax_id %s"%(tax_id))
        if sequence_type_id:
            extraArgumentList.append('--sequence_type_id %s'%(sequence_type_id))
        if chrOrder is not None:
            extraArgumentList.append("--chrOrder %s"%(chrOrder))
        job = self.addAbstractPlotJob(
            executable=executable,
            inputFileList=inputFileList,
            inputFile=inputFile,
            outputFile=outputFile,
            outputFnamePrefix=outputFnamePrefix,
            xColumnHeader=xColumnHeader, xColumnPlotLabel=xColumnPlotLabel,
            whichColumn=whichColumn,
            whichColumnHeader=whichColumnHeader,
            whichColumnPlotLabel=whichColumnPlotLabel,
            logX=logX, logY=logY,
            valueForNonPositiveYValue=valueForNonPositiveYValue,
            xScaleLog=xScaleLog, yScaleLog=yScaleLog,
            missingDataNotation=missingDataNotation,
            minNoOfTotal=minNoOfTotal, maxNoOfTotal=maxNoOfTotal,
            figureDPI=figureDPI, formatString=formatString,
            ylim_type=ylim_type,
            samplingRate=samplingRate, need_svg=need_svg,
            inputFileFormat=inputFileFormat,
            outputFileFormat=outputFileFormat,
            extraArguments=extraArguments,
            extraArgumentList=extraArgumentList,
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs,
            transferOutput=transferOutput,
            job_max_memory=job_max_memory,
            sshDBTunnel=sshDBTunnel,
            objectWithDBGenomeArguments=objectWithDBGenomeArguments,
            **keywords)
        return job

    def addAbstractPlotJob(self, executable=None,
        inputFileList=None, inputFile=None,
        outputFile=None,
        outputFnamePrefix=None,
        xColumnHeader=None, xColumnPlotLabel=None,
        whichColumn=None, whichColumnHeader=None,
        whichColumnPlotLabel=None,
        logX=None, logY=None, valueForNonPositiveYValue=-1,
        xScaleLog=0, yScaleLog=0,
        missingDataNotation='NA',
        title=None,
        minNoOfTotal=None, maxNoOfTotal=None,
        figureDPI=300, formatString='.', markerSize=None,
        ylim_type=1, samplingRate=1, legendType=None,
        need_svg=False,
        inputFileFormat=None, outputFileFormat=None,
        extraArguments=None,
        extraArgumentList=None,
        parentJob=None, parentJobLs=None,
        extraDependentInputLs=None, extraOutputLs=None,
        transferOutput=True,
        job_max_memory=200,
        sshDBTunnel=False, key2ObjectForJob=None,
        objectWithDBArguments=None, **keywords):
        """
        #no spaces or parenthesis or any other shell-vulnerable letters
        #  in the x or y axis labels (whichColumnPlotLabel, xColumnPlotLabel)
        Check AbstractMatrixFileWalker.py or AbstractPlot.py for updated arguments:

('outputFname', 0, ): [None, 'o', 1, 'output file for the figure.'],\
('minNoOfTotal', 1, int): [100, 'M', 1, 
    'minimum no of total variants (denominator of inconsistent rate)'],\
('title', 0, ): [None, 't', 1, 'title for the figure.'],\
('figureDPI', 1, int): [200, 'f', 1, 'dpi for the output figures (png)'],\
('formatString', 1, ): ['-', '', 1, 'formatString passed to matplotlib plot'],\
('ylim_type', 1, int): [1, 'y', 1, 'y-axis limit type, 1: whatever matplotlib decides. 2: min to max'],\
('samplingRate', 1, float): [1, 's', 1, 'how often you include the data, a probability between 0 and 1.'],\
('whichColumn', 0, int): [3, 'w', 1, 
    'data from this column (index starting from 0) is plotted as y-axis value'],\
('whichColumnHeader', 0, ): ["", 'W', 1, 
    'column header for the data to be plotted as y-axis value, substitute whichColumn'],\
('whichColumnPlotLabel', 0, ): ['', 'D', 1, 'plot label for data of the whichColumn', ],\
('logY', 0, int): [0, '', 1, 'value 0: nothing; 1: log(), 2: -log(). replacing self.logWhichColumn.'],\
('valueForNonPositiveYValue', 1, float): [50, '', 1, 
    'if the whichColumn value is not postive and logWhichColumn is on,\
        what yValue should be.'],\
('xColumnHeader', 1, ): ['', 'l', 1, 'header of the x-axis data column, ' ],\
('xColumnPlotLabel', 0, ): ['', 'x', 1, 'x-axis label (posColumn) in manhattan plot', ],\
('need_svg', 0, ): [0, 'n', 0, 'whether need svg output', ],\
('legendType', 0, int): [0, '', 1, '0: no legend; 1: legend'], \

inputFileFormat   1: csv-like plain text file; 2: YHPyTables.YHFile; 3: HDF5MatrixFile; . "1"(default)

        """
        if extraOutputLs is None:
            extraOutputLs = []
        if key2ObjectForJob is None:
            key2ObjectForJob = {}
        if executable is None:
            executable = self.AbstractPlot
        if extraDependentInputLs is None:
            extraDependentInputLs = []
        if inputFileList:
            extraDependentInputLs.extend(inputFileList)
        if extraArgumentList is None:
            extraArgumentList = []

        if outputFnamePrefix:
            extraArgumentList.append('--outputFnamePrefix %s'%(outputFnamePrefix))
            if outputFile is None:
                extraOutputLs.append(File('%s.png'%(outputFnamePrefix)))
                if need_svg:
                    extraOutputLs.append(File('%s.svg'%(outputFnamePrefix)))
        if minNoOfTotal is not None:
            extraArgumentList.append('--minNoOfTotal %s'%(minNoOfTotal))
        if maxNoOfTotal:
            extraArgumentList.append("--maxNoOfTotal %s"%(maxNoOfTotal))
        if figureDPI:
            extraArgumentList.append('--figureDPI %s'%(figureDPI))
        if formatString:
            extraArgumentList.append('--formatString %s'%(formatString))
        if markerSize:
            extraArgumentList.append("--markerSize %s"%(markerSize))
        if ylim_type:
            extraArgumentList.append('--ylim_type %s'%(ylim_type))
        if samplingRate is not None:
            extraArgumentList.append('--samplingRate %s'%(samplingRate))
        if legendType!=None:
            extraArgumentList.append("--legendType %s"%(legendType))

        if xColumnHeader:
            extraArgumentList.append('--xColumnHeader %s'%(xColumnHeader))
        if xColumnPlotLabel:
            extraArgumentList.append("--xColumnPlotLabel %s"%(xColumnPlotLabel))
        if whichColumnHeader:
            extraArgumentList.append("--whichColumnHeader %s"%(whichColumnHeader))
        if whichColumn:
            extraArgumentList.append("--whichColumn %s"%(whichColumn))
        if whichColumnPlotLabel:
            extraArgumentList.append("--whichColumnPlotLabel %s"%(whichColumnPlotLabel))
        if title:
            extraArgumentList.append("--title %s"%(title))
        if logX:
            extraArgumentList.append("--logX %s"%(logX))
        if logY:
            extraArgumentList.append('--logY %s'%(logY))
        if xScaleLog:
            extraArgumentList.append("--xScaleLog %s"%(xScaleLog))
        if yScaleLog:
            extraArgumentList.append("--yScaleLog %s"%(yScaleLog))

        if valueForNonPositiveYValue:
            extraArgumentList.append("--valueForNonPositiveYValue %s"%\
                (valueForNonPositiveYValue))
        if inputFileFormat:
            extraArgumentList.append("--inputFileFormat %s"%(inputFileFormat))
        if outputFileFormat:
            extraArgumentList.append("--outputFileFormat %s"%(outputFileFormat))
        if need_svg:
            extraArgumentList.append('--need_svg')
            if not outputFnamePrefix:
                outputFnamePrefix = os.path.splitext(outputFile.name)[0]
            extraOutputLs.append(File('%s.svg'%(outputFnamePrefix)))
        if extraArguments:
            extraArgumentList.append(extraArguments)

        job = self.addGenericJob(
            executable=executable,
            inputArgumentOption='-i',
            inputFile=inputFile, 
            outputArgumentOption='-o',
            outputFile=outputFile,
            inputFileList = inputFileList,
            extraArgumentList=extraArgumentList,
            parentJob=parentJob,
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs,
            extraOutputLs=extraOutputLs,
            transferOutput=transferOutput,
            key2ObjectForJob=key2ObjectForJob,
            job_max_memory=job_max_memory,
            sshDBTunnel=sshDBTunnel,
            objectWithDBArguments=objectWithDBArguments,
            **keywords)
        return job

    def addAbstractMatrixFileWalkerJob(self, executable=None,
        inputFileList=None,
        inputFile=None,
        outputFile=None,
        outputFnamePrefix=None,
        whichColumn=None, whichColumnHeader=None,
        logY=None, valueForNonPositiveYValue=-1,
        minNoOfTotal=10,
        samplingRate=1,
        inputFileFormat=None, outputFileFormat=None,
        extraArguments=None,
        extraArgumentList=None,
        parentJob=None, parentJobLs=None, extraOutputLs=None,
        extraDependentInputLs=None,
        transferOutput=True, job_max_memory=200,
        sshDBTunnel=False,
        objectWithDBArguments=None, **keywords):
        """
('outputFname', 0, ): [None, 'o', 1, 'output file for the figure.'],\
('minNoOfTotal', 1, int): [100, 'i', 1, 
    'minimum no of total variants (denominator of inconsistent rate)'],\
('samplingRate', 1, float): [1, 's', 1, 
    'how often you include the data, a probability between 0 and 1.'],\
('whichColumn', 0, int): [3, 'w', 1, 
    'data from this column (index starting from 0) is plotted as y-axis value'],\
('whichColumnHeader', 0, ): ["", 'W', 1, \
    'column header for the data to be plotted as y-axis value, substitute whichColumn'],\
('whichColumnPlotLabel', 0, ): ['', 'D', 1, 'plot label for data of the whichColumn', ],\
('logWhichColumn', 0, int): [0, 'g', 0, 'whether to take -log of the whichColumn'],\
('positiveLog', 0, int): [0, 'p', 0, 'toggle to take log, rather than -log(), \
    only effective when logWhichColumn is toggled. '],\
('valueForNonPositiveYValue', 1, float): [50, '', 1, 
    'if the whichColumn value is not postive and logWhichColumn is on,\
        what yValue should be.'],\

        """
        return self.addAbstractPlotJob(
            executable=executable, 
            inputFileList=inputFileList,
            inputFile=inputFile,
            outputFile=outputFile, 
            outputFnamePrefix=outputFnamePrefix,
            xColumnHeader=None, xColumnPlotLabel=None,
            whichColumn=whichColumn,
            whichColumnHeader=whichColumnHeader, whichColumnPlotLabel=None,
            logY=logY,
            valueForNonPositiveYValue=valueForNonPositiveYValue,
            missingDataNotation=None,
            minNoOfTotal=minNoOfTotal,
            figureDPI=None, formatString=None, ylim_type=None,
            samplingRate=samplingRate, need_svg=False,
            extraArguments=extraArguments,
            extraArgumentList=extraArgumentList,
            parentJob=parentJob, parentJobLs=parentJobLs,
            extraOutputLs=extraOutputLs,
            extraDependentInputLs=extraDependentInputLs,
            transferOutput=transferOutput, 
            job_max_memory=job_max_memory,
            sshDBTunnel=sshDBTunnel,
            objectWithDBArguments=objectWithDBArguments, **keywords)

    def addAbstractGenomeFileWalkerJob(self, executable=None,
        inputFileList=None,
        inputFile=None,
        outputFile=None,
        outputFnamePrefix=None,
        whichColumn=None, whichColumnHeader=None,
        logY=None, valueForNonPositiveYValue=-1,
        minNoOfTotal=10,
        samplingRate=1,
        chrColumnHeader=None,
        tax_id=60711, sequence_type_id=1, chrOrder=None,
        positionHeader=None,
        inputFileFormat=None, outputFileFormat=None,
        extraArguments=None,
        extraArgumentList=None,
        parentJob=None, parentJobLs=None,
        extraDependentInputLs=None,
        transferOutput=True, job_max_memory=200,
        sshDBTunnel=False,
        objectWithDBGenomeArguments=None, **keywords):
        """
        """
        if extraArgumentList is None:
            extraArgumentList=[]

        if objectWithDBGenomeArguments is None:
            objectWithDBGenomeArguments = self
        if chrColumnHeader is not None:
            extraArgumentList.append('--chromosomeHeader %s'%(chrColumnHeader))
        if tax_id:
            extraArgumentList.append("--tax_id %s"%(tax_id))
        if sequence_type_id:
            extraArgumentList.append('--sequence_type_id %s'%(sequence_type_id))
        if chrOrder is not None:
            extraArgumentList.append("--chrOrder %s"%(chrOrder))
        if positionHeader is not None:
            extraArgumentList.append('--positionHeader %s'%(positionHeader))
        return self.addAbstractPlotJob(
            executable=executable,
            inputFileList=inputFileList,
            inputFile=inputFile,
            outputFile=outputFile, 
            outputFnamePrefix=outputFnamePrefix,
            xColumnHeader=None, xColumnPlotLabel=None,
            whichColumn=whichColumn,
            whichColumnHeader=whichColumnHeader, whichColumnPlotLabel=None,
            logY=logY,
            valueForNonPositiveYValue=valueForNonPositiveYValue,
            missingDataNotation=None,
            minNoOfTotal=minNoOfTotal,
            figureDPI=None, formatString=None, ylim_type=None, 
            samplingRate=samplingRate, need_svg=False,
            extraArguments=extraArguments,
            extraArgumentList=extraArgumentList,
            parentJob=parentJob, parentJobLs=parentJobLs, 
            extraDependentInputLs=extraDependentInputLs,
            transferOutput=transferOutput,
            job_max_memory=job_max_memory,
            sshDBTunnel=sshDBTunnel, 
            objectWithDBGenomeArguments=objectWithDBGenomeArguments,
            **keywords)


    def addDrawHistogramJob(self, executable=None,
        inputFileList=None, inputFile=None,
        outputFile=None, outputFnamePrefix=None,
        title=None,
        whichColumn=None, whichColumnHeader=None, 
        whichColumnPlotLabel=None,
        xScaleLog=0, yScaleLog=0,
        logY=None,
        valueForNonPositiveYValue=-1,
        missingDataNotation='NA',
        minNoOfTotal=None, figureDPI=300,
        formatString=None, ylim_type=None,
        samplingRate=1,
        need_svg=False, legendType=None,
        logCount=False, inputFileFormat=None,
        extraArguments=None,
        extraArgumentList=None,
        parentJobLs=None,
        extraDependentInputLs=None,
        transferOutput=True, job_max_memory=200,
        **keywords):
        """
        #no spaces or parenthesis or any other shell-vulnerable letters
        #  in the x or y axis labels (whichColumnPlotLabel, xColumnPlotLabel)
('outputFname', 0, ): [None, 'o', 1, 'output file for the figure.'],\
('minNoOfTotal', 1, int): [100, 'i', 1, 
    'minimum no of total variants (denominator of inconsistent rate)'],\
('title', 0, ): [None, 't', 1, 'title for the figure.'],\
('figureDPI', 1, int): [200, 'f', 1, 'dpi for the output figures (png)'],\
('formatString', 1, ): ['-', '', 1, 'formatString passed to matplotlib plot'],\
('ylim_type', 1, int): [1, 'y', 1, 'y-axis limit type, 1: 0 to max. 2: min to max'],\
('samplingRate', 1, float): [1, 's', 1, 'how often you include the data, a probability between 0 and 1.'],\
('whichColumn', 0, int): [3, 'w', 1, 
    'data from this column (index starting from 0) is plotted as y-axis value'],\
('whichColumnHeader', 0, ): ["", 'W', 1, 
    'column header for the data to be plotted as y-axis value, substitute whichColumn'],\
('whichColumnPlotLabel', 0, ): ['', 'D', 1, 'plot label for data of the whichColumn', ],\
('logWhichColumn', 0, int): [0, 'g', 0, 'whether to take -log of the whichColumn'],\
('positiveLog', 0, int): [0, 'p', 0, 'toggle to take log, rather than -log(), \
    only effective when logWhichColumn is toggled. '],\
('valueForNonPositiveYValue', 1, float): [50, '', 1, 
    'if the whichColumn value is not postive and logWhichColumn is on,\
        what yValue should be.'],\
('need_svg', 0, ): [0, 'n', 0, 'whether need svg output', ],\

        """
        if extraArguments is None:
            extraArguments = ""
        if logCount:
            extraArguments += " --logCount "
        if executable is None:
            executable = self.DrawHistogram
        return self.addAbstractPlotJob(
            executable=executable,
            inputFileList=inputFileList,
            inputFile=inputFile,
            outputFile=outputFile, 
            outputFnamePrefix=outputFnamePrefix,
            title=title,
            xColumnHeader=None, xColumnPlotLabel=None,
            whichColumn=whichColumn,
            whichColumnHeader=whichColumnHeader,
            whichColumnPlotLabel=whichColumnPlotLabel,
            xScaleLog=xScaleLog, yScaleLog=yScaleLog,
            logY=logY,
            valueForNonPositiveYValue=valueForNonPositiveYValue,
            missingDataNotation=missingDataNotation,
            minNoOfTotal=minNoOfTotal,
            figureDPI=figureDPI, formatString=formatString,
            ylim_type=ylim_type,
            samplingRate=samplingRate, need_svg=need_svg,
            legendType=legendType,
            inputFileFormat=inputFileFormat,
            extraArguments=extraArguments,
            extraArgumentList=extraArgumentList,
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs,
            transferOutput=transferOutput, 
            job_max_memory=job_max_memory,
            **keywords)

    def addDraw2DHistogramOfMatrixJob(self, executable=None,
        inputFileList=None,
        inputFile=None,
        outputFile=None,
        outputFnamePrefix=None,
        xColumnHeader=None, xColumnPlotLabel=None,
        whichColumn=None,
        whichColumnHeader=None, whichColumnPlotLabel=None,
        logX=False, logY=False, logZ=False,
        valueForNonPositiveYValue=-1,
        missingDataNotation='NA',
        minNoOfTotal=100,
        figureDPI=300, formatString='.',
        samplingRate=1,
        need_svg=False,
        inputFileFormat=None, outputFileFormat=None,
        zColumnHeader=None,
        extraArguments=None,
        extraArgumentList=None,
        parentJobLs=None,
        extraDependentInputLs=None,
        transferOutput=True,
        job_max_memory=200, **keywords):
        """
        executable could be None, default is self.Draw2DHistogramOfMatrix

        """
        if extraArgumentList is None:
            extraArgumentList = []
        if zColumnHeader:
            extraArgumentList.append("--zColumnHeader %s"%(zColumnHeader))
        if logZ:
            extraArgumentList.append("--logZ %s"%(logZ))
        if executable is None:
            executable = self.Draw2DHistogramOfMatrix
        return self.addAbstractPlotJob(
            executable=executable, inputFileList=inputFileList,
            inputFile=inputFile,
            outputFile=outputFile,
            outputFnamePrefix=outputFnamePrefix,
            xColumnHeader=xColumnHeader, xColumnPlotLabel=xColumnPlotLabel,
            whichColumn=whichColumn,
            whichColumnHeader=whichColumnHeader,
            whichColumnPlotLabel=whichColumnPlotLabel,
            logX=logX, logY=logY,
            valueForNonPositiveYValue=valueForNonPositiveYValue,
            missingDataNotation=missingDataNotation,
            minNoOfTotal=minNoOfTotal,
            figureDPI=figureDPI, formatString=formatString,
            ylim_type=None,
            samplingRate=samplingRate,
            need_svg=need_svg,
            inputFileFormat=inputFileFormat,
            outputFileFormat=outputFileFormat,
            extraArguments=extraArguments,
            extraArgumentList=extraArgumentList,
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs,
            transferOutput=transferOutput,
            job_max_memory=job_max_memory,
            **keywords)

    def setupMoreOutputAccordingToSuffixAndNameTupleList(self,
        outputFnamePrefix=None,
        suffixAndNameTupleList=None, extraOutputLs=None, key2ObjectForJob=None):
        """
        split from addPlinkJob()
        """
        for suffixNameTuple in suffixAndNameTupleList:
            if len(suffixNameTuple)==1:
                suffix = suffixNameTuple[0]
                name = suffix[1:].replace('.', '_')
                #replace dot with underscore. as dot is used to access
                #  method/attribute of python object.
                # i.e. ".prune.in" is accessible as job.prune_inFile
            elif len(suffixNameTuple)>=2:
                suffix, name = suffixNameTuple[:2]
            outputFile = File('%s%s'%(outputFnamePrefix, suffix))
            extraOutputLs.append(outputFile)
            key2ObjectForJob['%sFile'%(name)] = outputFile

if __name__ == '__main__':
    from argparse import ArgumentParser
    ap = ArgumentParser()
    ap.add_argument("-i", "--input_path", type=str, required=True,
        help="the path to the input folder or file.")
    ap.add_argument("--inputSuffixList", type=str,
        help='Coma-separated list of input file suffices. Used to exclude input files.'
        'If None, no exclusion. The dot is part of the suffix, .tsv not tsv.'
        'Common zip suffices (.gz, .bz2, .zip, .bz) will be ignored in obtaining the suffix.')
    ap.add_argument("-F", "--pegasusFolderName", default='input',
        help='The path relative to the workflow running root. '
        'This folder will contain pegasus input & output. '
        'It will be created during the pegasus staging process. '
        'It is useful to separate multiple sub-workflows. '
        'If empty or None, everything is in the pegasus root.')
    
    ap.add_argument("-o", "--output_path", type=str, required=True,
        help="The path to the output file that will contain the Pegasus DAG.")
    
    ap.add_argument("--tmpDir", type=str, default='/tmp/',
        help='Default: %(default)s. A local folder for some jobs (MarkDup) to store temp data.'
        '/tmp/ can be too small sometimes.')
    ap.add_argument("--max_walltime", type=int, default=4320,
        help='Default: %(default)s. Maximum wall time for any job, in minutes. 4320=3 days.'
        'Used in addGenericJob(). Most clusters have upper limit for runtime.')
    ap.add_argument("--jvmVirtualByPhysicalMemoryRatio", type=float, default=1.2,
        help='Default: %(default)s. '
        'If a job virtual memory (usually 1.2X of JVM resident memory) exceeds request, '
        "it will be killed on some clusters. "
        "This will make sure your job requests enough memory.")
    
    ap.add_argument("-l", "--site_handler", required=True,
        help="The name of the computing site where the jobs run and executables are stored. "
        "Check your Pegasus configuration in submit.sh.")
    ap.add_argument("-j", "--input_site_handler", type=str,
        help="It is the name of the site that has all the input files."
        "Possible values can be 'local' or same as site_handler."
        "If not given, it is asssumed to be the same as site_handler and "
        "the input files will be symlinked into the running folder."
        "If input_site_handler=local, the input files will be transferred "
        "to the computing site by pegasus-transfer.")
    ap.add_argument("-C", "--cluster_size", type=int, default=30,
        help="Default: %(default)s. "
        "This number decides how many of pegasus jobs should be clustered into one job. "
        "Good if your workflow contains many quick jobs. "
        "It will reduce Pegasus monitor I/O.")
    
    ap.add_argument("--debug", action='store_true',
        help='Toggle debug mode.')
    ap.add_argument("--report", action='store_true',
        help="Toggle verbose mode. Default: %(default)s.")
    ap.add_argument("--needSSHDBTunnel", action='store_true',
        help="If all DB-interacting jobs need a ssh tunnel to "
        "access a database that is inaccessible to computing nodes.")
    args = ap.parse_args()
    instance = AbstractWorkflow(
        input_path=args.input_path,
        inputSuffixList=args.inputSuffixList, 
        pegasusFolderName=args.pegasusFolderName,
        output_path=args.output_path,
        
        tmpDir=args.tmpDir,
        max_walltime=args.max_walltime,
        jvmVirtualByPhysicalMemoryRatio=args.jvmVirtualByPhysicalMemoryRatio,
        
        site_handler=args.site_handler, 
        input_site_handler=args.input_site_handler,
        cluster_size=args.cluster_size,
        
        needSSHDBTunnel=args.needSSHDBTunnel,
        debug=args.debug,
        report=args.report)
    instance.setup_run()
    instance.end_run()



#!/usr/bin/env python3
"""
2012.11.20 data structures for IO
"""
import os, sys

import numpy
import networkx as nx
from palos.utils import PassingData
from palos.io.HDF5MatrixFile import HDF5MatrixFile, addAttributeDictToYHTableInHDF5Group
from palos.io.YHPyTables import YHTable, YHFile, castPyTablesRowIntoPassingData
from palos.polymorphism.SNP import getGenomeWideResultFromHDF5MatrixFile

import tables
from tables import UInt64Col, Float64Col, StringCol, Int64Col

class AssociationTable(tables.IsDescription):
	"""
	2012.12.18 pytable class to store the genome-wide association result
	"""
	id = UInt64Col(pos=0)
	locus_id = UInt64Col(pos=1, dflt=0)
	chromosome = StringCol(64, pos=2, dflt='')	#64 byte-long
	start = UInt64Col(pos=3, dflt=0)
	stop = UInt64Col(pos=4, dflt=0)
	score = Float64Col(pos=5, dflt=-1)
	mac = Int64Col(pos=6, dflt=-1)
	maf = Float64Col(pos=7, dflt=-1)
	genotype_var_perc = Float64Col(pos=8, dflt=-1)
	beta_list = Float64Col(shape=(5,), pos=9, dflt=-1)	#2013.1.9
	beta_pvalue_list = Float64Col(shape=(5,), pos=9, dflt=-1)

class AssociationTableFile(YHFile):
	"""
	Example to read an association table file:
		associationTableFile = AssociationTableFile(absPath, mode='r', autoRead=False)
		self.no_of_loci = associationTableFile.nrows # will NOT read in the whole data
		associationTableFile.close()
	
    	>>> associationTableFile.associationTable.shape
        (213655,)
        >>> associationTableFile.associationTable[:3]
        array([ (1L, 1L, '1', 657L, 657L, 0.00464357927467517, 46, 0.2658959537572254, -1.0, [-1.0, -1.0, -1.0, -1.0, -1.0], [-1.0, -1.0, -1.0, -1.0, -1.0]),
           (2L, 2L, '1', 3102L, 3102L, 0.008631413389472068, 50, 0.28901734104046245, -1.0, [-1.0, -1.0, -1.0, -1.0, -1.0], [-1.0, -1.0, -1.0, -1.0, -1.0]),
           (3L, 3L, '1', 4648L, 4648L, 0.0026632828069891286, 50, 0.28901734104046245, -1.0, [-1.0, -1.0, -1.0, -1.0, -1.0], [-1.0, -1.0, -1.0, -1.0, -1.0])], 
          dtype=[('id', '<u8'), ('locus_id', '<u8'), ('chromosome', 'S64'), ('start', '<u8'), ('stop', '<u8'), ('score', '<f8'), ('mac', '<i8'), ('maf', '<f8'), ('genotype_var_perc', '<f8'), ('beta_list', '<f8', (5,)), ('beta_pvalue_list', '<f8', (5,))])
        >>> associationTableFile.associationTable.colIDList
        ['id', 'locus_id', 'chromosome', 'start', 'stop', 'score', 'mac', 'maf', 'genotype_var_perc', 'beta_list', 'beta_pvalue_list']

		
	Example to write:
		associationTableFile = AssociationTableFile(output_fname, mode='w')
		attributeDict = None
		writer = gwr.outputInHDF5MatrixFile(tableObject=associationTableFile.associationTable, attributeDict=attributeDict)
		associationTableFile.close()
	"""
	def __init__(self, inputFname=None, mode='r', \
				tableName='association', groupNamePrefix='group', tableNamePrefix='table',\
				filters=None, expectedrows=300000, autoRead=True, autoWrite=True, \
				min_MAF=None, do_log10_transformation=False, **keywords):
		self.min_MAF = min_MAF
		self.genome_wide_result = None
		self.associationTable = None
		self.do_log10_transformation = do_log10_transformation
		
		YHFile.__init__(self, path=inputFname, mode=mode, \
			tableName=tableName, groupNamePrefix=groupNamePrefix, tableNamePrefix=tableNamePrefix,\
			rowDefinition=AssociationTable, filters=filters, expectedrows=expectedrows,\
			autoRead=autoRead, autoWrite=autoWrite,\
			debug=0, report=0, )
		

		"""
		if mode=='r' and self.readInData:
			self.associationTable = self.getTableObject(tableName=self.tableName)
			self._readInGWR(min_MAF=self.min_MAF, tableObject=self.associationTable)
		elif mode=='w':
			self.associationTable = self.createNewTable(tableName=self.tableName, rowDefinition=AssociationTable, \
											expectedrows=300000)
		"""
		self.associationTable = self.getTableObject(tableName=self.tableName)
	
	def _readInData(self, tableName=None, tableObject=None, do_log10_transformation=None):
		"""
		"""
		YHFile._readInData(self, tableName=tableName, tableObject=tableObject)
		
		if tableName is None:
			tableName = self.tableName
		if do_log10_transformation is None:
			do_log10_transformation = getattr(self, 'do_log10_transformation', False)
		pdata = PassingData(min_MAF=self.min_MAF)
		self.genome_wide_result = getGenomeWideResultFromHDF5MatrixFile(reader=self, tableName=tableName, tableObject=tableObject,\
							min_value_cutoff=None, do_log10_transformation=do_log10_transformation, pdata=pdata,\
							construct_chr_pos2index=False, construct_data_obj_id2index=False, \
							construct_locus_db_id2index=True,\
							report=True)
		return self.genome_wide_result


class LocusMapTable(tables.IsDescription):
	"""
	2013.1.9 a pytable to store the locus_id and (chromosome, start, stop)
		used by Association.py to find chr, start, stop for input loci
	"""
	id = UInt64Col(pos=0)
	locus_id = UInt64Col(pos=1, dflt=0)
	chromosome = StringCol(64, pos=2, dflt="")	#64 byte-long
	start = UInt64Col(pos=3, dflt=0)
	stop = UInt64Col(pos=4, dflt=0)

class LocusMapTableFile(YHFile):

	"""
	usage examples:
	
		locusMapTableFile = LocusMapTableFile(self.outputFname, mode='w')
		locusMapTableFile.addAttributeDict(attributeDict)
		locusMapTableFile.appendOneRow(pdata, cellType=2)
		
		#for read-only
		locusMapTableFile = LocusMapTableFile(inputFname, mode='r')
		locus_id2chr_pos = locusMapTableFile.locus_id2chr_pos
	"""
	def __init__(self, inputFname=None, mode='r', \
				tableName='locus_map', groupNamePrefix='group', tableNamePrefix='table',\
				filters=None, expectedrows=500000, autoRead=True, autoWrite=True, \
				**keywords):
		
		self.locus_id2chr_pos = None
		YHFile.__init__(self, path=inputFname, mode=mode, \
				tableName=tableName, groupNamePrefix=groupNamePrefix, tableNamePrefix=tableNamePrefix,\
				rowDefinition=LocusMapTable, filters=filters, expectedrows=expectedrows,\
				autoRead=autoRead, autoWrite=autoWrite,\
				debug=0, report=0, **keywords)
		
		
		#if (mode=='r' or mode == 'a')  and self.readInData:
		#	self.locusMapTable = self.getTableObject(tableName=self.tableName)
		#	self._readInMap(tableObject=self.locusMapTable)
		#elif mode == 'w':
		#	self.locusMapTable = self.createNewTable(tableName=self.tableName, rowDefinition=LocusMapTable,\
		#										expectedrows=500000)
		self.locusMapTable = self.getTableObject(tableName=self.tableName)
		
	def _readInData(self, tableName=None, tableObject=None):
		"""
		2012.1.9
		"""
		YHFile._readInData(self, tableName=tableName, tableObject=tableObject)
		
		if tableObject is None:
			tableObject = self.getTableObject(tableName=tableName)
		
		sys.stderr.write("Reading the locus map from HDF5 file %s ..."%(self.inputFname))
		"""
		for attributeName, value in self.getAttributes().items():
			HDF5AttributeNameLs.append(attributeName)
			setattr(, attributeName, value)
		"""
		counter = 0
		real_counter = 0
		self.locus_id2chr_pos = {}
		for row in tableObject:
			if not row['chromosome']:	#empty chromosome, which happens when inputFname contains no valid peaks, but the default null peak (only one).
				continue
			counter += 1
			chr_pos = (row['chromosome'], row['start'], row['stop'])
			locus_id = row['locus_id']
			if locus_id not in self.locus_id2chr_pos:
				self.locus_id2chr_pos[locus_id] = chr_pos
				real_counter += 1
			else:
				chr_pos = self.locus_id2chr_pos[locus_id]
				sys.stderr.write("Warning: locus_id %s is already in locus_id2chr_pos with chr,start,stop(%s, %s, %s).\n"%\
								(locus_id, chr_pos[0], chr_pos[1], chr_pos[2]))
		sys.stderr.write("%s loci (%s total) with unique locus_id.\n"%(real_counter, counter))
		return self.locus_id2chr_pos
	
	def getRowGivenLocusID(self, locus_id=None):
		"""
		2013.1.9
		"""
		query = self.where("""(locus_id ==%s) """%(locus_id))
		rowToReturn = None
		for row in query:
			rowToReturn = castPyTablesRowIntoPassingData(row)
		return rowToReturn
		#return (row['chromosome'], row['start'], row['stop'])

	def getRowGivenChrPos(self, chromosome=None, start=None, stop=None):
		"""
		2013.1.9
			(chromosome, start) might not be unique (some SNPs and CNVs have some overlap).
		"""
		query = self.where("""(chromosome==%s) & (start==%s) & (stop==%s) """%(chromosome, start, stop))
		rowToReturn = None
		for row in query:
			rowToReturn = castPyTablesRowIntoPassingData(row)
		return rowToReturn

def getAssociationLandscapeDataFromHDF5File(inputFname=None, associationTableName='association', \
										landscapeTableName='landscape', min_MAF=0.1):
	"""
	2012.11.20
		input is in HDF5MatrixFile format (which is output of variation/src/association_peak/DefineAssociationLandscape.py)
		contains two hdf5 groups. one is by associationTableName. the other is by landscapeTableName.
	"""
	pdata = PassingData(min_MAF=min_MAF)
	genome_wide_result = getGenomeWideResultFromHDF5MatrixFile(inputFname=inputFname, \
						min_value_cutoff=None, do_log10_transformation=False, pdata=pdata,\
						construct_chr_pos2index=False, construct_data_obj_id2index=False, \
						construct_locus_db_id2index=True,\
						report=True, tableName=associationTableName)
	
	returnData = PassingData(genome_wide_result=genome_wide_result)
	
	sys.stderr.write("Reading landscape from %s ..."%(inputFname))
	current_obj = None
	bridge_ls = []
	locusLandscapeNeighborGraph = nx.Graph()
	reader = HDF5MatrixFile(inputFname, mode='r')
	landscapeTableObject = reader.getTableObject(tableName=landscapeTableName)
	returnData.HDF5AttributeNameLs = []
	for attributeName, value in landscapeTableObject.getAttributes().items():
		returnData.HDF5AttributeNameLs.append(attributeName)
		setattr(returnData, attributeName, value)
	
	for row in landscapeTableObject:
		if row.start_locus_id==0:	#empty data. happens when inputFname contains no valid landscape, but one default null data point.
			continue
		start_locus_id = row.start_locus_id
		stop_locus_id = row.stop_locus_id
		no_of_loci = row.no_of_loci
		deltaX = row.deltaX
		
		start_obj = genome_wide_result.get_data_obj_by_locus_db_id(start_locus_id)
		stop_obj = genome_wide_result.get_data_obj_by_locus_db_id(stop_locus_id)
		
		bridge_ls.append([start_obj, stop_obj, no_of_loci, deltaX])
		
		source_index = start_obj.index
		#genome_wide_result.get_data_obj_index_by_locus_db_id(start_locus_id)
		target_index = stop_obj.index
		
		locusLandscapeNeighborGraph.add_edge(source_index, target_index, \
									weight=None)
		locusLandscapeNeighborGraph[source_index][target_index]['no_of_loci'] = no_of_loci
		locusLandscapeNeighborGraph[source_index][target_index]['deltaX'] = deltaX
		
	del reader
	sys.stderr.write("%s bridges.\n"%(len(bridge_ls)))
	returnData.bridge_ls = bridge_ls
	returnData.locusLandscapeNeighborGraph = locusLandscapeNeighborGraph
	return returnData

def outputAssociationLandscapeInHDF5(bridge_ls=None, outputFname=None, writer=None, closeFile=False, tableName='landscape',\
							attributeDict=None,):
	"""
	2012.11.18
	"""
	sys.stderr.write("Outputting the %s bridges from the landscape ..."%(len(bridge_ls)))
	#output the data_object.id in bridge_ls to outputFname
	#each number below is counting bytes, not bits
	rowDefinition = [('start_locus_id','i8'),('stop_locus_id', 'i8'), ('no_of_loci','i8'), ('deltaX', 'i8')]
	if writer:
		tableObject = writer.createNewTable(tableName=tableName, rowDefinition=rowDefinition)
	elif outputFname:
		writer = HDF5MatrixFile(outputFname, mode='w', rowDefinition=rowDefinition, tableName=tableName)
		tableObject = writer.getTableObject(tableName=tableName)
	else:
		sys.stderr.write("Error: no writer(%s) or filename(%s) to dump.\n"%(writer, outputFname))
		sys.exit(3)
	addAttributeDictToYHTableInHDF5Group(tableObject=tableObject, attributeDict=attributeDict)
	
	previous_locus_id = None
	cellList = []
	for bridge in bridge_ls:
		current_obj = bridge[0]
		obj_with_fastest_score_increase = bridge[1]
		no_of_loci, deltaX = bridge[2:4]
		dataTuple = (current_obj.db_id, obj_with_fastest_score_increase.db_id, no_of_loci, deltaX)
		cellList.append(dataTuple)
	tableObject.writeCellList(cellList)
	if closeFile:
		writer.close()
	sys.stderr.write("%s objects.\n"%(len(cellList)))
	return writer

def constructAssociationPeakRBDictFromHDF5File(inputFname=None, peakPadding=10000, tableName='association_peak'):
	"""
	2012.11.12
		similar to Stock_250kDB.constructRBDictFromResultPeak(), but from HDF5MatrixFile-like file
	"""
	from palos.algorithm.RBTree import RBDict
	from palos.polymorphism.CNV import CNVCompare, CNVSegmentBinarySearchTreeKey, get_overlap_ratio
	
	sys.stderr.write("Constructing association-peak RBDict from HDF5 file %s, (peakPadding=%s) ..."%(inputFname, peakPadding))
	reader = HDF5MatrixFile(inputFname, mode='r')
	associationPeakRBDict = RBDict()
	associationPeakRBDict.result_id = None	#2012.6.22
	associationPeakRBDict.peakPadding = peakPadding
	associationPeakRBDict.HDF5AttributeNameLs = []
	
	tableObject = reader.getTableObject(tableName=tableName)
	for attributeName, value in tableObject.getAttributes().items():
		associationPeakRBDict.HDF5AttributeNameLs.append(attributeName)
		setattr(associationPeakRBDict, attributeName, value)
	
	counter = 0
	real_counter = 0
	for row in tableObject:
		if not row.chromosome:	#empty chromosome, which happens when inputFname contains no valid peaks, but the default null peak (only one).
			continue
		counter += 1
		segmentKey = CNVSegmentBinarySearchTreeKey(chromosome=row.chromosome, \
						span_ls=[max(1, row.start - peakPadding), row.stop + peakPadding], \
						min_reciprocal_overlap=1, result_peak_id=None)
						#2010-8-17 overlapping keys are regarded as separate instances as long as they are not identical.
		if segmentKey not in associationPeakRBDict:
			associationPeakRBDict[segmentKey] = []
		else:
			sys.stderr.write("Warning: segmentKey of %s already in associationPeakRBDict with this row: %s.\n"%\
							(row, associationPeakRBDict[segmentKey][0]))
		associationPeakRBDict[segmentKey].append(row)
	sys.stderr.write("%s peaks in %s spans.\n"%(counter, len(associationPeakRBDict)))
	return associationPeakRBDict


def outputAssociationPeakInHDF5(association_peak_ls=None, filename=None, writer=None, tableName='association_peak', closeFile=True,\
							attributeDict=None,):
	"""
	2012.11.20
	"""
	sys.stderr.write("Dumping association peaks into %s (HDF5 format) ..."%(filename))
	#each number below is counting bytes, not bits
	rowDefinition = [('chromosome', HDF5MatrixFile.varLenStrType), ('start','i8'), ('stop', 'i8'), \
				('start_locus_id','i8'), ('stop_locus_id','i8'), \
				('no_of_loci', 'i8'), ('peak_locus_id', 'i8'), ('peak_score', 'f8')]
	if writer is None and filename:
		writer = HDF5MatrixFile(filename, mode='w', rowDefinition=rowDefinition, tableName=tableName)
		tableObject = writer.getTableObject(tableName=tableName)
	elif writer:
		tableObject = writer.createNewTable(tableName=tableName, rowDefinition=rowDefinition)
	else:
		sys.stderr.write("Error: no writer(%s) or filename(%s) to dump.\n"%(writer, filename))
		sys.exit(3)
	#add neighbor_distance, max_neighbor_distance, min_MAF, min_score, ground_score as attributes
	addAttributeDictToYHTableInHDF5Group(tableObject=tableObject, attributeDict=attributeDict)
	cellList = []
	#2012.11.28 sort it
	association_peak_ls.sort()
	for association_peak in association_peak_ls:
		dataTuple = (association_peak.chromosome, association_peak.start, association_peak.stop, \
					association_peak.start_locus_id, association_peak.stop_locus_id, \
					association_peak.no_of_loci,\
					association_peak.peak_locus_id, association_peak.peak_score)
		cellList.append(dataTuple)
	
	if tableObject is None:
		sys.stderr.write("Error: tableObject (name=%s) is None. could not write.\n"%(tableName))
		sys.exit(3)
	tableObject.writeCellList(cellList)
	if closeFile:
		writer.close()
	sys.stderr.write("%s objects.\n"%(len(cellList)))
	return writer


def constructAssociationLocusRBDictFromHDF5File(inputFname=None, locusPadding=0, tableName='association_locus'):
	"""
	2012.11.25
		similar to constructAssociationPeakRBDictFromHDF5File
	"""
	from palos.algorithm.RBTree import RBDict
	from palos.polymorphism.CNV import CNVCompare, CNVSegmentBinarySearchTreeKey, get_overlap_ratio
	
	sys.stderr.write("Constructing association-locus RBDict from HDF5 file %s, (locusPadding=%s) ..."%(inputFname, locusPadding))
	reader = HDF5MatrixFile(inputFname, mode='r')
	associationLocusRBDict = RBDict()
	associationLocusRBDict.locusPadding = locusPadding
	associationLocusRBDict.HDF5AttributeNameLs = []
	tableObject = reader.getTableObject(tableName=tableName)
	for attributeName, value in tableObject.getAttributes().items():
		associationLocusRBDict.HDF5AttributeNameLs.append(attributeName)
		setattr(associationLocusRBDict, attributeName, value)
	
	counter = 0
	real_counter = 0
	for row in tableObject:
		if not row.chromosome:	#empty chromosome, which happens when inputFname contains no valid locus, but the default null locus (only one).
			continue
		counter += 1
		segmentKey = CNVSegmentBinarySearchTreeKey(chromosome=row.chromosome, \
						span_ls=[max(1, row.start - locusPadding), row.stop + locusPadding], \
						min_reciprocal_overlap=1, no_of_peaks=row.no_of_peaks, \
						no_of_results=row.no_of_results, connectivity=row.connectivity)
						#2010-8-17 overlapping keys are regarded as separate instances as long as they are not identical.
		if segmentKey not in associationLocusRBDict:
			associationLocusRBDict[segmentKey] = []
		associationLocusRBDict[segmentKey].append(row)
	sys.stderr.write("%s peaks in %s spans.\n"%(counter, len(associationLocusRBDict)))
	return associationLocusRBDict

def outputAssociationLociInHDF5(associationLocusList=None, filename=None, writer=None, tableName='association_locus', \
					closeFile=True,\
					attributeDict=None):
	"""
	2012.12.10
		for each locus, output the association peaks that fall into the locus.
			for each association peak, include 
				* result-id 
				* phenotype id
				* chromosome
				* start
				* stop
				* start_locus
				* stop_locus
				* no_of_loci
				* peak_locus
				* peak-score
	2012.11.20
	"""
	sys.stderr.write("Dumping association loci into %s (HDF5 format) ..."%(filename))
	#each number below is counting bytes, not bits
	rowDefinition = [('chromosome', HDF5MatrixFile.varLenStrType), \
				('start','i8'), ('stop', 'i8'), \
				('no_of_peaks', 'i8'), ('connectivity', 'f8'), ('no_of_results', 'i8')]
	if writer is None and filename:
		writer = HDF5MatrixFile(filename, mode='w', rowDefinition=rowDefinition, tableName=tableName)
		tableObject = writer.getTableObject(tableName=tableName)
	elif writer:
		tableObject = writer.createNewTable(tableName=tableName, rowDefinition=rowDefinition)
	else:
		sys.stderr.write("Error: no writer(%s) or filename(%s) to dump.\n"%(writer, filename))
		sys.exit(3)
	#add neighbor_distance, max_neighbor_distance, min_MAF, min_score, ground_score as attributes
	addAttributeDictToYHTableInHDF5Group(tableObject=tableObject, attributeDict=attributeDict)
	cellList = []
	#2012.11.28 sort it
	associationLocusList.sort()
	for associationLocus in associationLocusList:
		dataTuple = (associationLocus.chromosome, associationLocus.start, associationLocus.stop, associationLocus.no_of_peaks,\
					associationLocus.connectivity, associationLocus.no_of_results)
		cellList.append(dataTuple)
	
	if tableObject is None:
		sys.stderr.write("Error: tableObject (name=%s) is None. could not write.\n"%(tableName))
		sys.exit(3)
	tableObject.writeCellList(cellList)
	if closeFile:
		writer.close()
	sys.stderr.write("%s objects.\n"%(len(cellList)))
	return writer
